from base64 import b64encode, b64decode
from datetime import datetime
from hashlib import md5
import json
import logging
from warnings import warn

import django
from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed

from .conf import settings
from .models import AgentSettings, Agent, SESSION_TOKEN_KEY


logger = logging.getLogger(__name__)


class AgentMiddleware(object):
    """
    This must be installed after
    :class:`~django.contrib.auth.middleware.AuthenticationMiddleware` to manage
    trusted agents.

    This middleware will set ``request.agent`` to an instance of
    :class:`django_agent_trust.models.Agent`. ``request.agent.is_trusted`` will
    tell you whether the user's agent has been trusted.
    """
    def __init__(self):
        if django.VERSION < (1, 4):
            warn('django_agent_trust requires Django 1.4 or higher')
            raise MiddlewareNotUsed()

    def process_request(self, request):
        if request.user.is_authenticated():
            AgentSettings.objects.get_or_create(user=request.user)

            request.agent = self._load_agent(request)
        else:
            request.agent = Agent.untrusted_agent(request.user)

        return None

    def process_response(self, request, response):
        agent = getattr(request, 'agent', None)

        if (agent is not None) and agent.user.is_authenticated():
            self._save_agent(agent, response)

        return response

    def _load_agent(self, request):
        cookie_name = self._cookie_name(self._get_username(request.user))
        max_age = self._max_cookie_age(request.user.agentsettings)

        # 'e30=' is base64 for '{}'
        encoded = request.get_signed_cookie(cookie_name, default='e30=',
                                            max_age=max_age)

        agent = self._decode_cookie(encoded, request.user)

        if (agent.session is not None) and (agent.session != request.session.get(SESSION_TOKEN_KEY)):
            agent = Agent.untrusted_agent(request.user)

        return agent

    def _decode_cookie(self, encoded, user):
        agent = None

        content = b64decode(encoded.encode('utf-8')).decode('utf-8')
        data = json.loads(content)

        logger.debug('Decoded agent: {0}'.format(data))

        if data.get('username') == self._get_username(user):
            agent = Agent.from_jsonable(data, user)
            if self._should_discard_agent(agent):
                agent = None

        if agent is None:
            agent = Agent.untrusted_agent(user)

        logger.debug('Loaded agent: username={0}, is_trusted={1}, trusted_at={2}, serial={3}'.format(
            self._get_username(user), agent.is_trusted, agent.trusted_at,
            agent.serial)
        )

        return agent

    def _should_discard_agent(self, agent):
        expiration = agent.trust_expiration
        if (expiration is not None) and (expiration < datetime.now()):
            return True

        if agent.serial < agent.user.agentsettings.serial:
            return True

        return False

    def _save_agent(self, agent, response):
        logger.debug('Saving agent: username={0}, is_trusted={1}, trusted_at={2}, serial={3}'.format(
            self._get_username(agent.user), agent.is_trusted, agent.trusted_at,
            agent.serial)
        )

        cookie_name = self._cookie_name(self._get_username(agent.user))
        encoded = self._encode_cookie(agent, agent.user)
        max_age = self._max_cookie_age(agent.user.agentsettings)

        response.set_signed_cookie(cookie_name, encoded, max_age=max_age,
                                   path=settings.AGENT_COOKIE_PATH,
                                   domain=settings.AGENT_COOKIE_DOMAIN,
                                   secure=settings.AGENT_COOKIE_SECURE,
                                   httponly=settings.AGENT_COOKIE_HTTPONLY)

    def _encode_cookie(self, agent, user):
        data = agent.to_jsonable()
        content = json.dumps(data)
        encoded = b64encode(content.encode('utf-8')).decode('utf-8')

        return encoded

    def _cookie_name(self, username):
        suffix = md5(username.encode('utf-8')).hexdigest()[16:]

        return '{0}-{1}'.format(settings.AGENT_COOKIE_NAME, suffix)

    def _max_cookie_age(self, agentsettings):
        """
        Returns the max cookie age based on inactivity limits.
        """
        days = settings.AGENT_INACTIVITY_DAYS

        try:
            int(days) * 86400
        except Exception:
            raise ImproperlyConfigured('AGENT_INACTIVITY_DAYS must be a number.')

        user_days = agentsettings.inactivity_days
        if (user_days is not None) and (user_days < days):
            days = user_days

        return days * 86400

    def _get_username(self, user):
        """
        Return the username of a user in a model- and version-indepenedent way.
        """
        return user.get_username() if hasattr(user, 'get_username') else user.username



from datetime import datetime, timedelta

import django
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.db import IntegrityError
from django.http import HttpResponse
from django.test import TestCase
from django.test.client import Client, RequestFactory
from django.utils import six
if django.VERSION < (1, 7):
    from django.utils import unittest
else:
    import unittest

from .conf import settings
from .decorators import trusted_agent_required
from .models import Agent, AgentSettings
from .middleware import AgentMiddleware


now = lambda: datetime.now().replace(microsecond=0)


@unittest.skipIf(django.VERSION < (1, 4), 'Requires Django 1.4')
class AgentTrustTestCase(TestCase):
    """
    Base class with some custom-user-aware utilities.
    """
    @classmethod
    def setUpClass(cls):
        super(AgentTrustTestCase, cls).setUpClass()

        try:
            from django.contrib.auth import get_user_model
        except ImportError:
            from django.contrib.auth.models import User
            cls.User = User
            cls.User.get_username = lambda self: self.username
            cls.USERNAME_FIELD = 'username'
        else:
            cls.User = get_user_model()
            cls.USERNAME_FIELD = cls.User.USERNAME_FIELD

    def create_user(self, username, password):
        """
        Try to create a user, honoring the custom user model, if any. This may
        raise an exception if the user model is too exotic for our purposes.
        """
        return self.User.objects.create_user(username, password=password)


class AgentCodingTestCase(AgentTrustTestCase):
    """
    Tests as much of the middleware as possible without the request/response
    cycle. Any tests that need to manipulate the date will be at this level.
    """
    def setUp(self):
        try:
            self.alice = self.create_user('alice', 'alice')
            AgentSettings.objects.create(user=self.alice)
            self.bob = self.create_user('bob', 'bob')
        except IntegrityError:
            self.skipTest("Unable to create a test user.")

        self.middleware = AgentMiddleware()

    @property
    def agentsettings(self):
        return self.alice.agentsettings

    def test_coverage(self):
        six.u(str(self.agentsettings))

    def test_trust_anonymous(self):
        with self.assertRaises(Exception):
            Agent.trusted_agent(AnonymousUser())

    def test_session_anonymous(self):
        with self.assertRaises(Exception):
            Agent.session_agent(AnonymousUser(), 0)

    def test_untrusted(self):
        agent = self._roundtrip_agent(Agent.untrusted_agent(self.alice))

        self.assertTrue(not agent.is_trusted)
        self.assertTrue(not agent.is_session)
        self.assertEqual(agent.trusted_at, None)
        self.assertEqual(agent.serial, -1)

    def test_trusted(self):
        trusted_at = now()

        agent = self._roundtrip(True, trusted_at, None, 1, None)

        self.assertTrue(agent.is_trusted)
        self.assertTrue(not agent.is_session)
        self.assertEqual(agent.trusted_at, trusted_at)
        self.assertEqual(agent.serial, 1)

    def test_expired_global_only(self):
        trusted_at = now() - timedelta(days=7)

        with settings(AGENT_TRUST_DAYS=5):
            agent = self._roundtrip(True, trusted_at, None, 1, None)

        self.assertTrue(not agent.is_trusted)

    def test_expired_user_only(self):
        trusted_at = now() - timedelta(days=7)
        self.agentsettings.trust_days = 5

        agent = self._roundtrip(True, trusted_at, None, 1, None)

        self.assertTrue(not agent.is_trusted)

    def test_expired_agent_only(self):
        trusted_at = now() - timedelta(days=7)

        agent = self._roundtrip(True, trusted_at, 5, 1, None)

        self.assertTrue(not agent.is_trusted)

    def test_expired_global_precedence(self):
        trusted_at = now() - timedelta(days=7)
        self.agentsettings.trust_days = 14

        with settings(AGENT_TRUST_DAYS=5):
            agent = self._roundtrip(True, trusted_at, 14, 1, None)

        self.assertTrue(not agent.is_trusted)

    def test_expired_user_precedence(self):
        trusted_at = now() - timedelta(days=7)
        self.agentsettings.trust_days = 5

        with settings(AGENT_TRUST_DAYS=14):
            agent = self._roundtrip(True, trusted_at, 14, 1, None)

        self.assertTrue(not agent.is_trusted)

    def test_expired_agent_precedence(self):
        trusted_at = now() - timedelta(days=7)
        self.agentsettings.trust_days = 14

        with settings(AGENT_TRUST_DAYS=14):
            agent = self._roundtrip(True, trusted_at, 5, 1, None)

        self.assertTrue(not agent.is_trusted)

    def test_expired_none(self):
        trusted_at = now() - timedelta(days=7)
        self.agentsettings.trust_days = 14

        with settings(AGENT_TRUST_DAYS=14):
            agent = self._roundtrip(True, trusted_at, 14, 1, None)

        self.assertTrue(agent.is_trusted)
        self.assertEqual(agent.trusted_at, trusted_at)
        self.assertEqual(agent.serial, 1)

    def test_revoked(self):
        trusted_at = now()
        self.agentsettings.serial = 2

        agent = self._roundtrip(True, trusted_at, None, 1, None)

        self.assertTrue(not agent.is_trusted)

    def test_session(self):
        agent = self._roundtrip(True, now(), None, 1, '1234')

        self.assertTrue(agent.is_trusted)
        self.assertTrue(agent.is_session)

    def test_cross_user(self):
        AgentSettings.objects.get_or_create(user=self.bob)

        agent = Agent.trusted_agent(self.alice)
        encoded = self.middleware._encode_cookie(agent, self.alice)
        agent = self.middleware._decode_cookie(encoded, self.bob)

        self.assertTrue(not agent.is_trusted)

    def test_inactivity_config(self):
        with self.assertRaises(ImproperlyConfigured):
            with settings(AGENT_INACTIVITY_DAYS=()):
                self.middleware._max_cookie_age(self.agentsettings)

    def test_inactivity_precedence(self):
        self.agentsettings.inactivity_days = 30

        self.assertEqual(self.middleware._max_cookie_age(self.agentsettings), 30 * 86400)

    def _roundtrip(self, *args, **kwargs):
        agent = Agent(self.alice, *args, **kwargs)

        return self._roundtrip_agent(agent)

    def _roundtrip_agent(self, agent):
        return self._decode_cookie(self._encode_cookie(agent))

    def _encode_cookie(self, agent):
        return self.middleware._encode_cookie(agent, self.alice)

    def _decode_cookie(self, encoded):
        return self.middleware._decode_cookie(encoded, self.alice)


class DecoratorTest(AgentTrustTestCase):
    def setUp(self):
        try:
            self.alice = self.create_user('alice', 'alice')
            AgentSettings.objects.create(user=self.alice)
        except IntegrityError:
            self.skipTest("Unable to create a test user.")

        self.factory = RequestFactory()

    def test_view_1_untrusted(self):
        request = self.factory.get('/')
        request.user = self.alice
        request.agent = Agent.untrusted_agent(self.alice)

        response = decorated_view_1(request)

        self.assertEqual(response.status_code, 302)

    def test_view_1_trusted(self):
        request = self.factory.get('/')
        request.user = self.alice
        request.agent = Agent.trusted_agent(self.alice)

        response = decorated_view_1(request)

        self.assertEqual(response.status_code, 200)

    def test_view_2_untrusted(self):
        request = self.factory.get('/')
        request.user = self.alice
        request.agent = Agent.untrusted_agent(self.alice)

        response = decorated_view_2(request)

        self.assertEqual(response.status_code, 302)

    def test_view_2_trusted(self):
        request = self.factory.get('/')
        request.user = self.alice
        request.agent = Agent.trusted_agent(self.alice)

        response = decorated_view_2(request)

        self.assertEqual(response.status_code, 200)


@trusted_agent_required
def decorated_view_1(request):
    return HttpResponse()


@trusted_agent_required()
def decorated_view_2(request):
    return HttpResponse()


class HttpTestCase(AgentTrustTestCase):
    """
    Tests that exercise the full request/response cycle. These are less
    precise, but touch more code.
    """
    urls = 'django_agent_trust.test.urls'

    def setUp(self):
        try:
            user = self.create_user('alice', 'alice')
            AgentSettings.objects.create(user=user)
            self.create_user('bob', 'bob')
        except IntegrityError:
            self.skipTest("Unable to create a test user.")
        else:
            self.alice = AgentClient('alice')
            self.bob = AgentClient('bob')

    def test_anonymous(self):
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 302)

    def test_anon_trust(self):
        self.alice.trust()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 302)

    def test_authenticated(self):
        self.alice.login()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 302)

    def test_trusted(self):
        self.alice.login()
        self.alice.trust()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 200)

    def test_anon(self):
        self.alice.trust()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 302)

    def test_revoked(self):
        self.alice.login()
        self.alice.trust()
        self.alice.revoke()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 302)

    def test_exotic_username(self):
        user = self.create_user('charlie@example.com', 'charlie')
        AgentSettings.objects.create(user=user)
        charlie = AgentClient('charlie@example.com', 'charlie')

        charlie.login()
        charlie.trust()
        response = charlie.get_restricted()

        self.assertEqual(response.status_code, 200)

    def test_persist(self):
        self.alice.login()
        self.alice.trust()
        self.alice.logout()
        self.alice.login()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 200)

    def test_trusted_session(self):
        self.alice.login()
        self.alice.trust_session()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 200)

    def test_old_session(self):
        self.alice.login()
        self.alice.trust_session()
        self.alice.logout()
        self.alice.login()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 302)

    def test_anon_session(self):
        self.alice.trust_session()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 302)

    def test_replace_trust(self):
        self.alice.login()
        self.alice.trust()
        self.alice.trust_session()
        self.alice.logout()
        self.alice.login()
        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 302)

    def test_other_trusted(self):
        self.alice.login()
        self.alice.trust()
        self.alice.logout()

        self.bob.login()
        response = self.bob.get_restricted()

        self.assertEqual(response.status_code, 302)

    def test_other_revoked(self):
        self.alice.login()
        self.alice.trust()

        self.bob.login()
        self.bob.trust()
        self.bob.revoke()
        self.bob.logout()

        response = self.alice.get_restricted()

        self.assertEqual(response.status_code, 200)

    def test_revoke_others(self):
        alice1 = AgentClient('alice')
        alice2 = AgentClient('alice')

        alice1.login()
        alice1.trust()

        alice2.login()
        alice2.trust()
        alice2.revoke_others()

        response = alice1.get_restricted()

        self.assertEqual(response.status_code, 302)


class AgentClient(Client):
    def __init__(self, username, password=None):
        super(AgentClient, self).__init__()

        self.username = username
        self.password = password if (password is not None) else username

    def login(self):
        return self.post('/login/', {'username': self.username, 'password': self.password})

    def logout(self):
        return self.post('/logout/')

    def get_restricted(self):
        return self.get('/restricted/')

    def trust(self):
        return self.post('/trust/')

    def trust_session(self):
        return self.post('/session/')

    def revoke(self):
        return self.post('/revoke/')

    def revoke_others(self):
        return self.post('/revoke_others/')



#!/usr/bin/env python

from setuptools import setup


setup(
    name='django-agent-trust',
    version='0.2.1',
    description="A framework for managing agent trust, such as public vs. private computers.",
    long_description=open('README').read(),
    author='Peter Sagerson',
    author_email='psagersDjwublJf@ignorare.net',
    packages=[
        'django_agent_trust',
        'django_agent_trust.migrations',
        'django_agent_trust.south_migrations',
        'django_agent_trust.test',
    ],
    url='https://bitbucket.org/psagers/django-agent-trust',
    license='BSD',
    install_requires=[
        'django>=1.4',
    ],
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Topic :: Security',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
